class TableAlreadyExists(Exception):
    """Exception is raised when we try to create a table 
       with a table name that already exists in Dynamodb.

    Attributes:
        table_name (str): Table name
        message (str): Explanation of the error
    """
    def __init__(self, table_name, message="Table already exists"):
        self.table_name = table_name
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} -> {self.table_name}'
    
class TableNotFound(Exception):
    """Exception is raised when we try to perform a Dynamodb 
       operation in a table that doesn't exist

    Attributes:
        table_name (str): Table name
        message (str): Explanation of the error
    """
    def __init__(self, table_name, message="Table doesn't exist"):
        self.table_name = table_name
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} -> {self.table_name}'

class ItemNotFound(Exception):
    """Exception is raised when the item is not available

    Attributes:
        table_name (str): Table name
        key (dict): Partition key,  Sort Key(Optional)
        message (str): Explanation of the error
    """
    def __init__(self, table_name, key, message="Item doesn't exist"):
        self.table_name = table_name
        self.key = key
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} -> Key: {self.key}, Table: {self.table_name}'

class QueryFilterValidationFailed(Exception):
    """Exception is raised when the Query filter is not valid

    Attributes:
        table_name (str): Table name
        message (str): Explanation of the error
    """
    def __init__(self, table_name, message="Item doesn't exist"):
        self.table_name = table_name
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message} -> Table: {self.table_name}'

class UnexpectedError(Exception):
    """Exception is raised when we perform an unexpected Dynamodb operation.

    Attributes:
        message (str): Explanation of the error
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'Unexpected Dynamodb operation -> {self.message}'


def create_attribute_names(attribute_names):
    """Create attribute names
    Args:
        attribute_names (str): Attribute names
    Returns:
        str: Expression attribute names
    """
    expression_attribute_names = {}
    for attribute_name in attribute_names:
        expression_attribute_names[f"#{attribute_name}"] = attribute_name
    return expression_attribute_names

def generate_expression_attribute_values(attributes_to_update, operation):
    """Generate Expression Attribute Values
    Args:
        attributes_to_update (dict): Attributes to update
        operation (str): Operation category
    Returns:
        update_expression (str): Describes all updates you want to perform on specified item
            Example: SET #domain_name = :value1, #owned_by = :value2
        expression_attribute_names (dict): Attribute name
            Example: {'#domain_name': 'domain_name', '#owned_by': 'owned_by'}
        expression_attribute_values (dict): Attribute values
            Example: {':value1': 'xbox.com', ':value2': 'Microsoft'}
    """
    update_expression = ""
    expression_attribute_names = {}
    expression_attribute_values = {}
    counter = 1
    for attribute_name, attribute_value in attributes_to_update.items():
        expression_attribute_names = create_attribute_names(attribute_name.split('.'))
        attribute_name = attribute_name.replace(".", ".#")
        if operation == "UPDATE_EXISTING_ATTRIBUTE_OR_ADD_NEW_ATTRIBUTE":
            if "SET" not in update_expression:
                update_expression = "SET "
            update_expression += f"#{attribute_name} = :value{counter}, "
        elif operation == "INCREASE_ATTRIBUTE_VALUE":
            if "SET" not in update_expression:
                update_expression = "SET "
            update_expression += f"#{attribute_name} = #{attribute_name} + :value{counter}, "
        elif operation == "ADD_ATTRIBUTE_TO_LIST":
            if "SET" not in update_expression:
                update_expression = "SET "
            update_expression += f"#{attribute_name} = list_append(#{attribute_name},:value{counter}), "
        elif operation == "ADD_ATTRIBUTE_TO_STRING_SET":
            if "ADD" not in update_expression:
                update_expression = "ADD "
            update_expression += f"#{attribute_name} :value{counter}, "
        elif operation == "DELETE_ATTRIBUTE_FROM_STRING_SET":
            if "DELETE" not in update_expression:
                update_expression = "DELETE "
            update_expression += f"#{attribute_name} :value{counter}, "
        if operation == "ADD_ATTRIBUTE_TO_LIST":
            expression_attribute_values[f":value{counter}"] = [attribute_value]
        elif operation == "ADD_ATTRIBUTE_TO_STRING_SET" or operation == "DELETE_ATTRIBUTE_FROM_STRING_SET":
            expression_attribute_values[f":value{counter}"] = set([attribute_value])
        else:
            expression_attribute_values[f":value{counter}"] = attribute_value
        counter = counter + 1
    update_expression = update_expression.rstrip(", ")
    return update_expression, expression_attribute_names, expression_attribute_values


from setuptools import setup, find_packages
import requests
import semantic_version

install_requires = [
    'boto3>=1.17.78',
    'botocore>=1.20.78'
]

def get_lucid_dynamodb_version():
    url = "https://pypi.org/pypi/LucidDynamodb/json"
    response = requests.request("GET", url, headers={}, data={})
    result = response.json()
    lucid_dynamodb_version = str(result.get("info").get("version"))
    current_version = semantic_version.Version(lucid_dynamodb_version)
    next_version = current_version.next_patch()
    return next_version

setup(
        name="LucidDynamodb",
        version=str(get_lucid_dynamodb_version()),
        author="Dinesh Sonachalam",
        author_email="dineshsonachalam@gmail.com",
        description="A simple Python wrapper to AWS Dynamodb",
        url="https://github.com/dineshsonachalam/Lucid-Dynamodb",
        long_description=open('README.md').read(),
        long_description_content_type='text/markdown',
        zip_safe=False,
        license='MIT',
        keywords='python dynamodb amazon',
        python_requires=">=3.1",
        install_requires=install_requires,
        packages=find_packages(),
        classifiers=[
                "Programming Language :: Python :: 3",
                "License :: OSI Approved :: MIT License",
                "Operating System :: OS Independent",
        ]
)


import boto3
from botocore.exceptions import ClientError
from LucidDynamodb.utils import generate_expression_attribute_values
from LucidDynamodb.exceptions import (
    TableAlreadyExists,
    TableNotFound,
    ItemNotFound,
    QueryFilterValidationFailed,
    UnexpectedError
)

class DynamoDb:
    def __init__(self, region_name=None, aws_access_key_id=None, aws_secret_access_key=None):
        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        if(self.region_name!=None or self.aws_access_key_id!=None or self.aws_secret_access_key!=None):
            self.db = boto3.resource(
                "dynamodb",
                region_name = region_name,
                aws_access_key_id = aws_access_key_id,
                aws_secret_access_key = aws_secret_access_key,
            )
        else:
            self.db = boto3.resource("dynamodb")

    def create_table(self, table_name, key_schema, attribute_definitions, provisioned_throughput, global_secondary_indexes=None):
        """Create a new table

        Args:
            table_name (str): Table name
            key_schema (list): A key schema specifies the attributes that make up the Partition key,  Sort Key(Optional)) of a table.
            attribute_definitions (list): An array of attributes that describe the key schema for the table.
            global_secondary_indexes (list, optional): An index with a partition key and a sort key that can be different from those on the base table.
            provisioned_throughput (dict): Provisioned throughput settings for this specified table.

        Returns:
            bool: Table creation is successful or failed
        """
        try:
            if(len(global_secondary_indexes)>0):
                table = self.db.create_table(
                    TableName=table_name,
                    KeySchema=key_schema,
                    AttributeDefinitions=attribute_definitions,
                    GlobalSecondaryIndexes=global_secondary_indexes,
                    ProvisionedThroughput=provisioned_throughput
                )
            else:
                table = self.db.create_table(
                    TableName=table_name,
                    KeySchema=key_schema,
                    AttributeDefinitions=attribute_definitions,
                    ProvisionedThroughput=provisioned_throughput
                )
                # Wait until the table exists.
                table.meta.client.get_waiter('table_exists').wait(TableName=table_name)
            return True
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceInUseException':
                raise TableAlreadyExists(table_name)
            else:
                raise UnexpectedError(e)

    def delete_table(self, table_name):
        """Delete a table

        Args:
            table_name (str): Table name

        Returns:
            bool: Table deletion is successful or failed
        """
        try:
            table = self.db.Table(table_name)
            table.delete()
            table.wait_until_not_exists()
            return True
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                raise TableNotFound(table_name)
            else:
                raise UnexpectedError(e)

    def read_all_table_names(self):
        """Get all table names

        Returns:
            list: List of table names
        """
        try:
            if(self.region_name!=None or self.aws_access_key_id!=None or self.aws_secret_access_key!=None):
                db_client = boto3.client(
                    "dynamodb",
                    region_name = self.region_name,
                    aws_access_key_id = self.aws_access_key_id,
                    aws_secret_access_key = self.aws_secret_access_key,
                )
            else:
                db_client = boto3.client(
                    "dynamodb"
                )
            table_names = db_client.list_tables()['TableNames']
            return table_names
        except ClientError as e:
            raise UnexpectedError(e)

    def create_item(self, table_name ,item):
        """Create a new Item

        Args:
            table_name (str): Table name
            item (dict): Item with Partition key,  Sort Key(Optional)

        Returns:
            bool: Item creation is successful or failed
        """
        try:
            table = self.db.Table(table_name)
            table.put_item(Item=item)
            return True
        except ClientError as e:
            raise UnexpectedError(e)

    def delete_item(self, table_name, key, condition_expression = "", expression_attribute_values=None):
        """Delete an Item

        Args:
            table_name (str): Table name
            key (dict): Partition key,  Sort Key(Optional)
            condition_expression (str, optional): condition_expression to prevent the item from being deleted if the condition is not met.
            expression_attribute_values (dict, optional): Expression attribute values.

        Returns:
            bool: Item deletion is successful or failed
        """
        try:
            table = self.db.Table(table_name)
            if(len(condition_expression)>0 and len(expression_attribute_values)>0):
                table.delete_item(
                    Key=key,
                    ConditionExpression=condition_expression,
                    ExpressionAttributeValues=expression_attribute_values
                )
            else:
                table.delete_item(
                    Key=key
                )
            return True
        except ClientError as e:
            raise UnexpectedError(e)

    def read_item(self, table_name, key):
        """Read an Item

        Args:
            table_name (str): Table name
            key (dict): Partition key,  Sort Key(Optional)

        Returns:
            dict: Item
        """
        try:
            table = self.db.Table(table_name)
            response = table.get_item(Key=key)
            item = response.get('Item')
            if item is not None:
                return item
            else:
                return ItemNotFound(table_name, key)
        except ClientError as e:
            raise UnexpectedError(e)

    def read_items_by_filter(self, table_name, key_condition_expression, global_secondary_index_name=None):
        """Read items by filter
           The Query operation will return all of the items from the table or index with that partition key value.

        Args:
            table_name (str): Table name
            key_condition_expression (boto3.dynamodb.conditions.Equals): Use the KeyConditionExpression parameter to
                provide a specific value for the partition key. You can optionally narrow the scope of the Query
                operation by specifying a sort key value and a comparison operator in KeyConditionExpression.
            global_secondary_index_name (str, optional): Name of the GlobalSecondaryIndex. Defaults to None.

        Returns:
            list: Table items
        """
        try:
            table = self.db.Table(table_name)
            if global_secondary_index_name != None:
                response = table.query(
                    IndexName=global_secondary_index_name,
                    KeyConditionExpression=key_condition_expression
                )
            else:
                response = table.query(
                    KeyConditionExpression=key_condition_expression
                )
            return response.get('Items')
        except ClientError as e:
            if e.response['Error']['Code'] == 'ValidationException':
                raise QueryFilterValidationFailed(table_name)
            else:
                raise UnexpectedError(e)

    def update_item(self, table_name, key,
                    attributes_to_update, operation="UPDATE_EXISTING_ATTRIBUTE_OR_ADD_NEW_ATTRIBUTE"):
        """Update an item
        Args:
            table_name (str): Table name
            key (dict): Partition key,  Sort Key(Optional)
            attributes_to_update (dict): Attributes data with K:V
            operation (str, optional): Update operation category
                Defaults to UPDATE_EXISTING_ATTRIBUTE_OR_ADD_NEW_ATTRIBUTE.
        
        Returns:
            bool: Attribute update is successful or failed
        """
        try:
            table = self.db.Table(table_name)
            update_expression, \
            expression_attribute_names, \
            expression_attribute_values = generate_expression_attribute_values(attributes_to_update, operation)
            if(len(update_expression)>0 and len(expression_attribute_names)>0 \
               and len(expression_attribute_values)>0):
                table.update_item(
                                Key=key,
                                UpdateExpression=update_expression,
                                ExpressionAttributeNames=expression_attribute_names,
                                ExpressionAttributeValues=expression_attribute_values,
                                ReturnValues="ALL_NEW"
                )
                return True
            else:
                return False
        except ClientError as e:
            raise UnexpectedError(e)

    def increase_attribute_value(self, table_name, key, attribute_name, increment_value):
        """Increase an existing attribute value

        Args:
            table_name (str): Table name
            key (dict): Partition key,  Sort Key(Optional)
            attribute_name (str): Name of the attribute
            increment_value (int): Increment value for an attribute

        Returns:
            bool: Attribute value is incremented or not
        """
        try:
            table = self.db.Table(table_name)
            attributes_to_update = {
                attribute_name: increment_value
            }
            operation = "INCREASE_ATTRIBUTE_VALUE"
            update_expression, expression_attribute_names, \
            expression_attribute_values = generate_expression_attribute_values(attributes_to_update, operation)
            if(len(update_expression)>0 and len(expression_attribute_names)>0 \
               and len(expression_attribute_values)>0):
                table.update_item(
                                Key=key,
                                UpdateExpression=update_expression,
                                ExpressionAttributeNames=expression_attribute_names,
                                ExpressionAttributeValues=expression_attribute_values,
                                ReturnValues="ALL_NEW"
                )
                return True
            else:
                return False
        except ClientError as e:
            raise UnexpectedError(e)

    def delete_attribute(self, table_name, key, attribute_name):
        """Delete an attribute from an item

        Args:
            table_name (str): Table name
            key (dict): Partition key,  Sort Key(Optional)
            attribute_name (str): Name of the Attribute

        Returns:
            bool: Attribute deletion is successful or failed
        """
        try:
            table = self.db.Table(table_name)
            table.update_item(
                                Key=key,
                                UpdateExpression=f"REMOVE {attribute_name}",
                                ReturnValues="ALL_NEW"
            )
            return True
        except ClientError as e:
            raise UnexpectedError(e)



import os
import uuid
from boto3.dynamodb.conditions import Key
from LucidDynamodb import DynamoDb
from LucidDynamodb.exceptions import (
    TableAlreadyExists,
    TableNotFound,
    ItemNotFound,
    QueryFilterValidationFailed,
    UnexpectedError
)

AWS_ACCESS_KEY_ID = os.getenv("AWS_ACCESS_KEY_ID")
AWS_SECRET_ACCESS_KEY = os.getenv("AWS_SECRET_ACCESS_KEY")
ITEM1_PARTITION_KEY = str(uuid.uuid4())

table_schema = {
	"TableName": "dev_jobs_test",
	"KeySchema": [
        {
            "AttributeName": "company_name",
            "KeyType": "HASH"
	    },
        {
            "AttributeName": "role_id",
            "KeyType": "RANGE"
	    }
    ],
	"AttributeDefinitions": [
        {
            "AttributeName": "company_name",
            "AttributeType": "S"
	    },
        {
            "AttributeName": "role_id",
            "AttributeType": "S"
	    }
     ],
	"GlobalSecondaryIndexes": [],
	"ProvisionedThroughput": {
		"ReadCapacityUnits": 1,
		"WriteCapacityUnits": 1
	}
}

db = DynamoDb(region_name="us-east-1",
              aws_access_key_id=AWS_ACCESS_KEY_ID,
              aws_secret_access_key=AWS_SECRET_ACCESS_KEY)

def test_create_new_table():
    try:
        table_creation_status = db.create_table(
                                            table_name=table_schema.get("TableName"),
                                            key_schema=table_schema.get("KeySchema"),
                                            attribute_definitions=table_schema.get("AttributeDefinitions"),
                                            global_secondary_indexes=table_schema.get("GlobalSecondaryIndexes"),
                                            provisioned_throughput=table_schema.get("ProvisionedThroughput")
        )
        assert table_creation_status == True
        db.create_table(
            table_name=table_schema.get("TableName"),
            key_schema=table_schema.get("KeySchema"),
            attribute_definitions=table_schema.get("AttributeDefinitions"),
            global_secondary_indexes=table_schema.get("GlobalSecondaryIndexes"),
            provisioned_throughput=table_schema.get("ProvisionedThroughput")
        )
    except TableAlreadyExists:
        assert True
        
def test_get_all_table_name():
    table_names = db.read_all_table_names()
    assert len(table_names)>0

def test_create_new_item():
    try:
        item_creation_status = db.create_item(
            table_name=table_schema.get("TableName"),
            item={
                "company_name": "Google",
                "role_id": ITEM1_PARTITION_KEY,
                "role": "Software Engineer 1",
                "salary": "$1,50,531",
                "locations": ["Mountain View, California", "Austin, Texas", "Chicago, IL"],
                "yearly_hike_percent": 8,
                "benefits": set(["Internet, Medical, Edu reimbursements",
                                "Health insurance",
                                "Travel reimbursements"
                                ]),
                "overall_review":{
                    "overall_rating" : "4/5",
                    "compensation_and_benefits": "3.9/5"
                }
            }
        )
        assert item_creation_status == True
        item_creation_status = db.create_item(
            table_name="TEST_TABLE",
            item={
                "company_name": "Google",
                "role_id": ITEM1_PARTITION_KEY,
                "role": "Software Engineer 1",
                "salary": "$1,50,531",
                "locations": ["Mountain View, California", "Austin, Texas", "Chicago, IL"],
                "yearly_hike_percent": 8,
                "benefits": set(["Internet, Medical, Edu reimbursements",
                                "Health insurance",
                                "Travel reimbursements"
                                ]),
                "overall_review":{
                    "overall_rating" : "4/5",
                    "compensation_and_benefits": "3.9/5"
                }
            }
        )
    except UnexpectedError:
        assert True

def test_read_item():
    try:
        item = db.read_item(
            table_name=table_schema.get("TableName"),
            key={
                "company_name": "Google",
                "role_id": ITEM1_PARTITION_KEY
            }
        )
        assert item != None
        item = db.read_item(
            table_name=table_schema.get("TableName"),
            key={
                "company_name": "Airbnb",
                "role_id": ITEM1_PARTITION_KEY
            }
        )
    except ItemNotFound:
        assert True

def test_increase_attribute_value():
    increase_attribute_status = db.increase_attribute_value(
        table_name=table_schema.get("TableName"),
        key={
            "company_name": "Google",
            "role_id": ITEM1_PARTITION_KEY
        },
        attribute_name="yearly_hike_percent",
        increment_value=5
    )
    assert increase_attribute_status==True

def test_update_existing_attribute():
    item_update_status = db.update_item(
        table_name=table_schema.get("TableName"),
        key={
            "company_name": "Google",
            "role_id": ITEM1_PARTITION_KEY
        },
        attributes_to_update={
            'role': 'Staff Software Engineer 2'
        }
    )
    assert item_update_status == True

def test_add_new_attribute():
    item_update_status = db.update_item(
        table_name=table_schema.get("TableName"),
        key={
            "company_name": "Google",
            "role_id": ITEM1_PARTITION_KEY
        },
        attributes_to_update={
            'overall_review.yearly_bonus_percent': 12
        }
    )
    assert item_update_status == True
    
def test_add_attribute_to_list():
    item_update_status = db.update_item(
        table_name=table_schema.get("TableName"),
        key={
            "company_name": "Google",
            "role_id": ITEM1_PARTITION_KEY
        },
        attributes_to_update={
            'locations': "Detroit, Michigan"
        },
        operation="ADD_ATTRIBUTE_TO_LIST"
    )
    assert item_update_status == True

def test_add_attributes_to_string_set():
    item_update_status = db.update_item(
        table_name=table_schema.get("TableName"),
        key={
            "company_name": "Google",
            "role_id": ITEM1_PARTITION_KEY
        },
        attributes_to_update={
            'benefits': "Free Food"
        },
        operation="ADD_ATTRIBUTE_TO_STRING_SET"
    )
    assert item_update_status == True

def test_delete_attribute_from_string_set():
    item_update_status = db.update_item(
        table_name=table_schema.get("TableName"),
        key={
            "company_name": "Google",
            "role_id": ITEM1_PARTITION_KEY
        },
        attributes_to_update={
            'benefits': "Free Food"
        },
        operation="DELETE_ATTRIBUTE_FROM_STRING_SET"
    )
    assert item_update_status == True

def test_delete_attribute_from_item():
    attribute_delete_status = db.delete_attribute(
        table_name=table_schema.get("TableName"),
        key={
              "company_name": "Google",
              "role_id": ITEM1_PARTITION_KEY
        },
        attribute_name="yearly_hike_percent")
    assert attribute_delete_status == True

def test_read_items_by_filter():
    try:
        item_creation_status = db.create_item(
            table_name=table_schema.get("TableName"),
            item={
                "company_name": "Google",
                "role_id": str(uuid.uuid4()),
                "role": "Software Architect",
                "salary": "$4,80,000",
                "locations": ["Mountain View, California"],
                "yearly_hike_percent": 13,
                "benefits": set(["Internet reimbursements"]),
                "overall_review":{
                    "overall_rating" : "3/5",
                    "compensation_and_benefits": "4.2/5"
                }
            }
        )
        assert item_creation_status == True
        items = db.read_items_by_filter(
                        table_name=table_schema.get("TableName"),
                        key_condition_expression=Key("company_name").eq("Google")
        )
        assert len(items)>0
        db.read_items_by_filter(
                        table_name=table_schema.get("TableName"),
                        key_condition_expression=Key("company_name").eq("Airbnb")
        )
    except QueryFilterValidationFailed:
        assert True

def test_delete_item():
    delete_item_status = db.delete_item(
        table_name=table_schema.get("TableName"),
        key={
            "company_name": "Google",
            "role_id": ITEM1_PARTITION_KEY
        }
    )
    assert delete_item_status == True

def test_delete_table():
    try:
        delete_table_status = db.delete_table(table_name=table_schema.get("TableName"))
        assert delete_table_status == True
    except TableNotFound:
        assert True